"""
This library maintains its own versioning
"""

VERSION = '0.0.1'

DEFAULT_AZURE_VERSION = '2024-12-01-preview'
DEFAULT_FUNCTION_MODEL = 'gpt-4o-mini'
DEFAULT_CHAT_MODEL = 'gpt-4o-mini'
DEFAULT_COMPLETION_MODEL = 'gpt-4o-mini'
DEFAULT_EMBEDDING_MODEL = 'text-embedding-3-small'