import importlib

from typing import Any
from lzl.logging import logger, null_logger

logger.set_module_name('lzl.ext.temporal.loop', 'tmprl', is_relative = True)
logger.set_module_name('lzl.ext.temporal', 'temporal', is_relative = True)
